import type {APIGatewayProxyEvent, APIGatewayProxyResult} from 'aws-lambda'
import {URL} from 'node:url'
import {match} from 'path-to-regexp'
import type {
  RestApiRouteHandler,
  RestPathParams,
  RestQueryParams,
  RestRequestBody,
  RestRequestHeaders,
  RestRequireAuthentication,
  RestResponseBody,
} from '../types'

const DEFAULT_RESPONSE_HEADERS = {
  'Content-Type': 'application/json; charset=utf-8',
  'Access-Control-Allow-Origin': '*',
  'Access-Control-Allow-Methods': '*',
  'Access-Control-Allow-Headers': '*',
} as const

export type AnyRestApiRouteHandler = RestApiRouteHandler<
  RestPathParams,
  RestQueryParams,
  RestRequestHeaders,
  RestRequestBody,
  RestResponseBody,
  RestRequireAuthentication
>

type MatchingRouteHandler = {
  routeHandler: AnyRestApiRouteHandler
  pathParams: RestPathParams
}

export class RestApi {
  private routeHandlers: AnyRestApiRouteHandler[] = []

  addRouteHandler<
    PathParams extends RestPathParams,
    QueryParams extends RestQueryParams,
    RequestHeaders extends RestRequestHeaders,
    RequestBody extends RestRequestBody,
    ResponseBody extends RestResponseBody,
    RequireAuthentication extends RestRequireAuthentication,
  >(
    routeHandler: RestApiRouteHandler<
      PathParams,
      QueryParams,
      RequestHeaders,
      RequestBody,
      ResponseBody,
      RequireAuthentication
    >,
  ) {
    this.routeHandlers = [...this.routeHandlers, routeHandler as unknown as AnyRestApiRouteHandler]
    return this
  }

  async handleRequest(
    proxyEvent: APIGatewayProxyEvent,
    onError?: (error: unknown) => Promise<void>,
  ): Promise<APIGatewayProxyResult> {
    try {
      const url = new URL(`${proxyEvent.headers?.['origin'] || 'https://no-origin'}${proxyEvent.path}`)
      const routeHandler = this.getMatchingRouteHandler(url, proxyEvent.httpMethod)
      if (!routeHandler && proxyEvent.httpMethod === 'OPTIONS') {
        const allowedHttpMethods = this.getMatchingRouteHandlers(url).map(({routeHandler}) => routeHandler.route.method)
        return {
          statusCode: 200,
          body: '',
          headers: {
            ...DEFAULT_RESPONSE_HEADERS,
            'Access-Control-Allow-Methods': allowedHttpMethods.join(','),
          },
        }
      }

      if (!routeHandler)
        return {
          statusCode: 405,
          headers: DEFAULT_RESPONSE_HEADERS,
          body: `Method ${proxyEvent.httpMethod} not allowed on ${proxyEvent.path}`,
        }

      const {
        routeHandler: {requestHandler, route},
        pathParams,
      } = routeHandler
      const userId = proxyEvent.requestContext.authorizer?.principalId
      if (route.requireAuthentication && !userId) return {statusCode: 401, body: ''}

      const responseBody = await requestHandler({
        pathParams,
        queryParams: this.extractQueryParams(proxyEvent),
        headers: proxyEvent.headers as Record<string, string>,
        body: proxyEvent.body && JSON.parse(proxyEvent.body),
        userId,
      })
      return {
        statusCode: responseBody.status,
        body: responseBody.body ? JSON.stringify(responseBody.body) : '',
        headers: {
          ...DEFAULT_RESPONSE_HEADERS,
          ...(responseBody.headers || {}),
        },
      }
    } catch (error) {
      if (onError) await onError(error)

      console.error(error)
      return {
        statusCode: 500,
        headers: DEFAULT_RESPONSE_HEADERS,
        body: JSON.stringify({error}),
      }
    }
  }

  private getMatchingRouteHandlers(url: URL): MatchingRouteHandler[] {
    return this.routeHandlers.reduce((matchingRouteHandlers, routeHandler) => {
      const pathParams = this.extractPathParams(routeHandler, url)
      const matchingRouteHandler: MatchingRouteHandler = {
        routeHandler,
        pathParams,
      }
      return pathParams ? [...matchingRouteHandlers, matchingRouteHandler] : matchingRouteHandlers
    }, [] as MatchingRouteHandler[])
  }

  private getMatchingRouteHandler(url: URL, httpMethod: string): MatchingRouteHandler | undefined {
    return this.getMatchingRouteHandlers(url).find(({routeHandler}) => routeHandler.route.method === httpMethod)
  }

  private extractPathParams(routeHandler: AnyRestApiRouteHandler, url: URL): object | never {
    const matcher = match(routeHandler.route.path, {decode: decodeURIComponent})
    return (matcher(url.pathname) || null)?.params as object | never
  }

  private extractQueryParams(proxyEvent: APIGatewayProxyEvent): object {
    const keys = new Set([
      ...Object.keys(proxyEvent.queryStringParameters || {}),
      ...Object.keys(proxyEvent.multiValueQueryStringParameters || {}),
    ])

    const queryParams: {[key: string]: string | string[] | undefined} = {}
    for (const key of keys.values()) {
      const singleValue = proxyEvent.queryStringParameters?.[key]
      const multiValues = proxyEvent.multiValueQueryStringParameters?.[key] || []
      queryParams[key] = multiValues.length > 1 ? multiValues : singleValue
    }
    return queryParams
  }
}
