import {useCallback, useMemo, useRef, useState} from 'react'
import type {WsApiAction, WsApiRequest, WsApiResponse, WsApiRoute} from '../types'

type WsApiConnection<Action extends WsApiAction, Request extends WsApiRequest<Action>> = {
  disconnect: () => void
  sendRequest: (request: Request) => void
}

export type WsApiOptions<
  Action extends WsApiAction,
  Request extends WsApiRequest<Action>,
  Response extends WsApiResponse,
> = {
  onConnected?: (connection: WsApiConnection<Action, Request>) => void
  onResponse?: (response: Response, connection: WsApiConnection<Action, Request>) => void
  onError?: (error: unknown, connection: WsApiConnection<Action, Request>) => void
  onDisconnected?: () => void
}
export const useWsApi = <
  Action extends WsApiAction,
  Request extends WsApiRequest<Action>,
  Response extends WsApiResponse,
>(
  wsApiUrl: string,
  route: WsApiRoute<Action, Request, Response> | undefined,
) => {
  const [error, setError] = useState<boolean>(false)
  const [connected, setConnected] = useState<boolean>(false)
  const webSocket = useRef<WebSocket>()

  const sendRequest = useCallback(
    (request: Omit<Request, 'action'>) => {
      if (!route) throw 'Not route'
      if (!webSocket.current) throw 'Not connected'

      webSocket.current.send(
        JSON.stringify({
          ...request,
          action: route.action,
        }),
      )
    },
    [route],
  )
  const disconnect = useCallback(() => webSocket.current?.close(), [])
  const connection: WsApiConnection<Action, Request> = useMemo(() => {
    return {
      disconnect,
      sendRequest,
    }
  }, [disconnect, sendRequest])

  const connect = useCallback(
    async (options: WsApiOptions<Action, Request, Response>) => {
      return new Promise<void>((resolve) => {
        if (
          webSocket.current &&
          (webSocket.current.readyState === WebSocket.CONNECTING || webSocket.current.readyState === WebSocket.OPEN)
        )
          throw 'Already connected'

        webSocket.current = new WebSocket(wsApiUrl)

        webSocket.current.onopen = () => {
          setConnected(true)
          setError(false)
          options?.onConnected?.(connection)
        }

        webSocket.current.onmessage = (ev) => {
          if (!ev.data || ev.data === '{}') return
          setError(false)
          options?.onResponse?.(JSON.parse(ev.data as string), connection)
        }

        webSocket.current.onerror = (error) => {
          setError(false)
          options?.onError?.(error, connection)
        }

        webSocket.current.onclose = () => {
          setConnected(false)
          options?.onDisconnected?.()
          webSocket.current = undefined
          resolve()
        }
      })
    },
    [wsApiUrl, connection],
  )

  return {
    connect,
    connected,
    connection,
    error,
    disconnect,
  }
}
