import type {APIGatewayProxyEvent, APIGatewayProxyResult} from 'aws-lambda'
import type {
  WsApiAction,
  WsApiConnection,
  WsApiRequest,
  WsApiRequestHandler,
  WsApiResponse,
  WsApiRouteHandler,
} from '../types'
import {closeConnection} from './closeConnection'
import {sendResponse} from './sendResponse'

export class WsApi {
  // eslint-disable-next-line @typescript-eslint/no-explicit-any
  private readonly routeHandlers: {[action: WsApiAction]: WsApiRequestHandler<any, any, any>} = {
    $connect: () => Promise.resolve(),
    $disconnect: () => Promise.resolve(),
  }

  addRouteHandler<Action extends WsApiAction, Request extends WsApiRequest<Action>, Response extends WsApiResponse>(
    routeHandler: WsApiRouteHandler<Action, Request, Response>,
  ) {
    this.routeHandlers[routeHandler.route.action] = routeHandler.requestHandler
    return this
  }

  async handleRequest(event: APIGatewayProxyEvent): Promise<APIGatewayProxyResult> {
    const {connectionId, domainName, stage, eventType} = event.requestContext

    if (!eventType) throw 'No eventType'
    if (!connectionId) throw 'No connectionId'
    if (!domainName) throw 'No domainName'

    let request: WsApiRequest<WsApiAction>
    switch (eventType) {
      case 'CONNECT':
        request = {action: '$connect'}
        break

      case 'DISCONNECT':
        request = {action: '$disconnect'}
        break

      case 'MESSAGE':
        if (!event.body) throw 'No body'
        request = JSON.parse(event.body) as WsApiRequest<WsApiAction>
        break

      default:
        throw `Unknown eventType ${eventType}`
    }

    const endpoint = `https://${domainName}/${stage}`
    const connection: WsApiConnection = {endpoint, connectionId}

    const requestHandler = this.routeHandlers[request.action]
    if (!requestHandler) {
      await closeConnection(connection)
      throw `No requestHandler found for action ${request.action}`
    }

    await requestHandler(request, {
      connection,
      sendResponse: async (response) => await sendResponse(connection, response),
      closeConnection: async () => await closeConnection(connection),
    })

    return {
      statusCode: 200,
      body: '',
    }
  }
}
